import pathlib
import time
import numpy as np
from scipy.sparse import csr_matrix, load_npz, triu
from ..cool import write_coo, get_chrom_offsets
import cooler
import pandas as pd
import h5py
from concurrent.futures import ProcessPoolExecutor, as_completed

"""
Matrix names

"""


def merge_cells_for_single_chromosome(output_dir,
                                      output_prefix,
                                      merge_type='E'):
    """
    Merge cell's E and T matrix to group matrices:
    E: Matrix normalized by global diagonal backgrounds, calculated from loop_bkg
    E2: E^2 of T, used to calculate global p values
    T: Matrix normalized by global diagonal and local backgrounds, then minus E (T is the delta matrix),
    calculated from loop_bkg
    T2: T^2 of T, used to calculate t test p values
    """
    start_time = time.time()
    if merge_type.upper() == 'E':
        print('Merging E (global diag norm), E2 (E^2 for global t-test).')
        # get cell paths
        cell_paths = [str(p) for p in pathlib.Path(output_dir).glob('*.E.npz')]
        n_cells = len(cell_paths)
        # get n_dims
        matrix = load_npz(cell_paths[0])
        n_dims = matrix.shape[0]
        # initialize
        e_sum = csr_matrix((n_dims, n_dims), dtype=np.float32)
        e2_sum = csr_matrix((n_dims, n_dims), dtype=np.float32)
        for i, chrom_path in enumerate(cell_paths):
            matrix = load_npz(chrom_path)
            e_sum += matrix
            e2_sum += matrix.multiply(matrix)
        write_coo(f'{output_prefix}.E.hdf', e_sum, chunk_size=None)
        write_coo(f'{output_prefix}.E2.hdf', e2_sum, chunk_size=None)
    else:
        print('Merging T (global and local norm) and T2 (T^2, for t-test) matrix.')
        # get cell paths
        cell_paths = [str(p) for p in pathlib.Path(output_dir).glob('*.T.npz')]
        n_cells = len(cell_paths)
        # get n_dims
        matrix = load_npz(cell_paths[0])
        n_dims = matrix.shape[0]
        # initialize
        t_sum = csr_matrix((n_dims, n_dims), dtype=np.float32)
        t2_sum = csr_matrix((n_dims, n_dims), dtype=np.float32)
        for i, chrom_path in enumerate(cell_paths):
            matrix = load_npz(chrom_path)
            t_sum += matrix
            t2_sum += matrix.multiply(matrix)
        write_coo(f'{output_prefix}.T.hdf', t_sum, chunk_size=None)
        write_coo(f'{output_prefix}.T2.hdf', t2_sum, chunk_size=None)
    print(f'Merge {n_cells} cells took {time.time() - start_time:.0f} seconds')
    return


def read_single_cool_chrom(cool_path, chrom):
    cool = cooler.Cooler(str(cool_path))
    matrix = triu(cool.matrix(balance=False, sparse=True).fetch(chrom))
    return matrix


def chrom_sum_iterator(chunk_dirs,
                       chrom_sizes,
                       chrom_offset,
                       matrix_type,
                       total_cells):
    print(f'Reading matrix {matrix_type}')
    for chrom in chrom_sizes.keys():
        # sum together multiple chunks
        # first
        cool_path = chunk_dirs[0] / f'{matrix_type}.cool'
        matrix = read_single_cool_chrom(cool_path, chrom)
        # others
        for chunk_dir in chunk_dirs[1:]:
            cool_path = chunk_dir / f'{matrix_type}.cool'
            matrix += read_single_cool_chrom(cool_path, chrom)

        matrix = matrix.tocoo()
        pixel_df = pd.DataFrame({
            'bin1_id': matrix.row,
            'bin2_id': matrix.col,
            'count': matrix.data
        })
        pixel_df.iloc[:, :2] += chrom_offset[chrom]
        pixel_df.iloc[:, -1] /= total_cells
        yield pixel_df


def save_single_matrix_type(cooler_path,
                            bins_df,
                            chunk_dirs,
                            chrom_sizes,
                            chrom_offset,
                            matrix_type,
                            total_cells):
    chrom_iter = chrom_sum_iterator(chunk_dirs,
                                    chrom_sizes,
                                    chrom_offset,
                                    matrix_type,
                                    total_cells)
    cooler.create_cooler(cool_uri=cooler_path,
                         bins=bins_df,
                         pixels=chrom_iter,
                         ordered=True,
                         dtypes={'count': np.float32})
    with h5py.File(cooler_path, 'a') as f:
        f.attrs['group_n_cells'] = total_cells
    return


def merge_group_chunks_to_group_cools(chrom_size_path,
                                      resolution,
                                      group,
                                      output_dir):
    # determine chunk dirs for the group:
    output_dir = pathlib.Path(output_dir).absolute()
    group_dir = output_dir / group
    group_dir.mkdir(exist_ok=True)

    chunk_dirs = list(output_dir.glob(f'{group}_chunk*'))

    # count total cells
    total_cells = 0
    for chunk_dir in chunk_dirs:
        total_cells += pd.read_csv(chunk_dir / 'cell_table.csv', index_col=0).shape[0]

    chrom_sizes = cooler.util.read_chromsizes(chrom_size_path)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    matrix_types = ['E', 'E2', 'T', 'T2', 'Q']
    with ProcessPoolExecutor(5) as exe:
        futures = {}
        for matrix_type in matrix_types:
            cooler_path = str(group_dir / f'{group}.{matrix_type}.cool')
            future = exe.submit(save_single_matrix_type,
                                cooler_path=cooler_path,
                                bins_df=bins_df,
                                chunk_dirs=chunk_dirs,
                                chrom_sizes=chrom_sizes,
                                chrom_offset=chrom_offset,
                                matrix_type=matrix_type,
                                total_cells=total_cells)
            futures[future] = matrix_type
        for future in as_completed(futures):
            matrix_type = futures[future]
            print(f'Matrix {matrix_type} generated')
            future.result()
    return



import argparse
import inspect
import logging
import sys
from .__main__ import setup_logging

log = logging.getLogger()

DESCRIPTION = """
hic-internal is used for automation, not intend to be used by end user. 
Use hicluster instead. 
"""

EPILOG = ''


def impute_chromosome_internal_subparser(subparser):
    parser = subparser.add_parser('impute-chromosome',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="RWR imputation for one chromosome in one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--scool_url",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True
    )

    parser.add_argument(
        "--logscale",
        type=bool,
        default=True,
    )

    parser.add_argument(
        "--pad",
        type=int,
        default=1
    )

    parser.add_argument(
        "--std",
        type=int,
        default=1
    )

    parser.add_argument(
        "--rp",
        type=float,
        default=0.5
    )

    parser.add_argument(
        "--tol",
        type=float,
        default=0.01
    )

    parser.add_argument(
        "--window_size",
        type=int,
        default=500000000
    )

    parser.add_argument(
        "--step_size",
        type=int,
        default=10000000
    )

    parser.add_argument(
        "--output_dist",
        type=int,
        default=500000000
    )

    parser.add_argument(
        "--min_cutoff",
        type=float,
        default=0
    )
    return


def aggregate_chromosomes_internal_subparser(subparser):
    parser = subparser.add_parser('aggregate-chromosomes',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Aggregate chromosome HDFs for one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--input_dir",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom_wildcard",
        type=str,
        default='{chrom}.hdf'
    )


def calculate_loop_matrix_internal_subparser(subparser):
    parser = subparser.add_parser('calculate-loop-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Calculate Loop Matrix E and T for single chromosome of one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--cell_url",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser.add_argument(
        "--dist",
        type=int,
        default=10050000
    )

    parser.add_argument(
        "--cap",
        type=int,
        default=5
    )

    parser.add_argument(
        "--pad",
        type=int,
        default=5
    )

    parser.add_argument(
        "--gap",
        type=int,
        default=2
    )

    parser.add_argument(
        "--min_cutoff",
        type=float,
        default=1e-6
    )

    parser.add_argument('--log_e', dest='log_e', action='store_true',
                        help='Normalize E at log scale')
    parser.set_defaults(log_e=False)


def merge_cell_impute_matrix_internal_subparser(subparser):
    parser = subparser.add_parser('merge-cell-impute-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Merge Loop Matrix E and T from cells to group")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--cell_urls_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )
    return


def merge_cell_loop_bkg_internal_subparser(subparser):
    parser = subparser.add_parser('merge-loop-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Merge Loop Matrix E and T from cells to group")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--merge_type",
        type=str,
        required=True
    )


def merge_group_chunks_internal_subparser(subparser):
    parser = subparser.add_parser('merge-group-chunks',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Aggregate group chunks cool files to one scool file per group.")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--group",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True
    )


def call_loop_internal_subparser(subparser):
    parser = subparser.add_parser('call-loop',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Call loop using group matrix stored in scool file.")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--group_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--thres_bl",
        type=float,
        default=1.33
    )

    parser_req.add_argument(
        "--thres_donut",
        type=float,
        default=1.33
    )

    parser_req.add_argument(
        "--thres_h",
        type=float,
        default=1.2
    )

    parser_req.add_argument(
        "--thres_v",
        type=float,
        default=1.2
    )

    parser_req.add_argument(
        "--fdr_thres",
        type=float,
        default=0.1
    )

    parser_req.add_argument(
        "--dist_thres",
        type=float,
        default=20000
    )

    parser_req.add_argument(
        "--size_thres",
        type=float,
        default=1
    )


def internal_main():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     epilog=EPILOG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    subparsers = parser.add_subparsers(
        title="functions",
        dest="command",
        metavar=""
    )

    # add subparsers
    current_module = sys.modules[__name__]
    # get all functions in parser
    for name, register_subparser_func in inspect.getmembers(current_module, inspect.isfunction):
        if 'internal_subparser' in name:
            register_subparser_func(subparsers)

    # initiate
    args = None
    if len(sys.argv) > 1:
        args = parser.parse_args()
    else:
        # print out help
        parser.parse_args(["-h"])
        exit()

    # set up logging
    if not logging.root.handlers:
        setup_logging(stdout=True,
                      quiet=False)

    # execute command
    args_vars = vars(args)
    for k, v in args_vars.items():
        logging.debug(k, v, type(v), sep='\t')

    cur_command = args_vars.pop('command')
    # Do real import here:
    if cur_command == 'impute-chromosome':
        from .impute.impute_chromosome import impute_chromosome as func
    elif cur_command == 'aggregate-chromosomes':
        from .cool.utilities import aggregate_chromosomes as func
    elif cur_command == 'calculate-loop-matrix':
        from .loop.loop_bkg import calculate_chrom_background_normalization as func
    elif cur_command == 'merge-loop-matrix':
        from .loop.merge_cell_to_group import merge_cells_for_single_chromosome as func
    elif cur_command == 'merge-group-chunks':
        from .loop.merge_cell_to_group import merge_group_chunks_to_group_cools as func
    elif cur_command == 'merge-cell-impute-matrix':
        from .impute.merge_cell_to_group import merge_cells_for_single_chromosome as func
    elif cur_command == 'call-loop':
        from .loop.loop_calling import call_loops as func
    else:
        log.debug(f'{cur_command} not Known, check the main function if else part')
        parser.parse_args(["-h"])
        return

    # run the command
    func(**args_vars)
    return



# command time python /gale/ddn/snm3C/humanPFC/code/embed_concatcell_chr.py --cell_list /gale/ddn/snm3C/humanPFC/smoothed_matrix/filelist/imputelist_pad1_std1_rp0.5_sqrtvc_chr${c}.txt --outprefix /gale/ddn/snm3C/humanPFC/smoothed_matrix/${res0}b_resolution/merged/pad1_std1_rp0.5_sqrtvc_chr${c} --res ${res}

import time
import h5py
import argparse
import numpy as np
from scipy.sparse import load_npz, save_npz, csr_matrix, vstack
from sklearn.decomposition import TruncatedSVD

def embed_concatcell_chr(cell_list, outprefix, res, dist=10000000, save_raw=True, dim=50):

	celllist = np.loadtxt(cell_list, dtype=np.str)

	with h5py.File(celllist[0], 'r') as f:
		ngene = f['Matrix'].attrs['shape'][0]
	idx = np.triu_indices(ngene, k=1)
	idxfilter = np.array([(yy - xx) < (dist / res + 1) for xx,yy in zip(idx[0], idx[1])])
	idx = (idx[0][idxfilter], idx[1][idxfilter])

	start_time = time.time()
	# matrix = np.zeros((len(celllist), np.sum(idxfilter)))
	matrix = []
	for i,cell in enumerate(celllist):
		with h5py.File(cell, 'r') as f:
			g = f['Matrix']
			A = csr_matrix((g['data'][()], g['indices'][()], g['indptr'][()]), g.attrs['shape'])
		# matrix[i] = A[idx]
		matrix.append(csr_matrix(A[idx]))
		if i%100==0:
			print(i, 'cells loaded', time.time() - start_time, 'seconds')

	matrix = vstack(matrix)

	if save_raw:
		save_npz(f'{outprefix}.npz', matrix)

	scalefactor = 100000
	matrix.data = matrix.data * scalefactor
	svd = TruncatedSVD(n_components=dim, algorithm='arpack')
	matrix_reduce = svd.fit_transform(matrix)
	matrix_reduce = np.concatenate((svd.singular_values_[None,:], matrix_reduce), axis=0)
	np.save(f'{outprefix}.svd{dim}.npy', matrix_reduce)
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--cell_list', type=str, default=None, help='Full path of a file containing the full path to all imputed matrices to be concatenate')
parser.add_argument('--outprefix', type=str, default=None, help='Prefix of concatenated matrix including directory')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer to generate contact matrix')
parser.add_argument('--dist', type=int, default=10000000, help='Maximum distance threshold of contacts to use')
parser.add_argument('--skip_raw', dest='save_raw', action='store_false', help='Not to save cell-by-feature matrix before SVD')
parser.set_defaults(save_raw=True)
parser.add_argument('--dim', type=int, default=50, help='Number of dimensions to return from SVD')
opt = parser.parse_args()

embed_concatcell_chr(opt.cell_list, opt.outprefix, opt.res, opt.dist, opt.save_raw, opt.dim)
'''



import pathlib
import pandas as pd
import schicluster

PACKAGE_DIR = pathlib.Path(schicluster.__path__[0])

with open(PACKAGE_DIR / 'loop/generate_matrix_chunk.Snakefile') as tmp:
    GENERATE_MATRIX_CHUNK_TEMPLATE = tmp.read()
with open(PACKAGE_DIR / 'loop/generate_matrix_group.Snakefile') as tmp:
    GENERATE_MATRIX_SCOOL_TEMPLATE = tmp.read()


def prepare_dir(output_dir, chunk_df, dist, cap, pad, gap, resolution,
                min_cutoff, chrom_size_path, keep_cell_matrix, log_e_str):
    output_dir.mkdir(exist_ok=True)
    cell_table_path = str((output_dir / 'cell_table.csv').absolute())
    chunk_df[['cell_url']].to_csv(cell_table_path)
    parameters = dict(dist=dist,
                      cap=cap,
                      pad=pad,
                      gap=gap,
                      resolution=resolution,
                      min_cutoff=min_cutoff,
                      cell_table_path=f'"{cell_table_path}"',
                      chrom_size_path=f'"{chrom_size_path}"',
                      keep_cell_matrix=keep_cell_matrix,
                      log_e_str=f'"{log_e_str}"')
    parameters_str = '\n'.join(f'{k} = {v}'
                               for k, v in parameters.items())

    with open(output_dir / 'Snakefile', 'w') as f:
        f.write(parameters_str + GENERATE_MATRIX_CHUNK_TEMPLATE)
    return


def prepare_loop_snakemake(cell_table_path, output_dir, chrom_size_path, genome, chunk_size=100, dist=10050000,
                           cap=5, pad=5, gap=2, resolution=10000, min_cutoff=1e-6,
                           keep_cell_matrix=False, cpu_per_job=10, log_e=False):
    cell_table = pd.read_csv(cell_table_path, index_col=0, sep='\t',
                             names=['cell_id', 'cell_url', 'cell_group'])
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    if log_e:
        log_e_str = '--log_e'
    else:
        log_e_str = ''

    chunk_parameters = dict(
        dist=dist,
        cap=cap,
        pad=pad,
        gap=gap,
        resolution=resolution,
        min_cutoff=min_cutoff,
        chrom_size_path=chrom_size_path,
        keep_cell_matrix=keep_cell_matrix,
        log_e_str=log_e_str
    )

    total_chunk_dirs = []
    group_chunks = {}
    for group, group_df in cell_table.groupby('cell_group'):
        group_chunks[group] = []
        if group_df.shape[0] <= chunk_size:
            this_dir = output_dir / f'{group}_chunk0'
            prepare_dir(this_dir, group_df, **chunk_parameters)
            total_chunk_dirs.append(this_dir)
            group_chunks[group].append(this_dir)
        else:
            group_df['chunk'] = [i // chunk_size for i in range(group_df.shape[0])]
            for chunk, chunk_df in group_df.groupby('chunk'):
                this_dir = output_dir / f'{group}_chunk{chunk}'
                prepare_dir(this_dir, chunk_df, **chunk_parameters)
                total_chunk_dirs.append(this_dir)
                group_chunks[group].append(this_dir)

    with open(output_dir / 'snakemake_cmd_step1.txt', 'w') as f:
        for chunk_dir in total_chunk_dirs:
            cmd = f'snakemake -d {chunk_dir} --snakefile {chunk_dir}/Snakefile -j {cpu_per_job}'
            f.write(cmd + '\n')

    # prepare the second step that merge cell chunks into groups
    scool_parameters = dict(
        output_dir=f'"{output_dir}"',
        chrom_size_path=f'"{chrom_size_path}"',
        resolution=resolution,
        genome=f'"{genome}"'
    )
    parameters_str = '\n'.join(f'{k} = {v}'
                               for k, v in scool_parameters.items())
    with open(output_dir / 'Snakefile', 'w') as f:
        f.write(parameters_str + GENERATE_MATRIX_SCOOL_TEMPLATE)
    with open(output_dir / 'snakemake_cmd_step2.txt', 'w') as f:
        cmd = f'snakemake -d {output_dir} --snakefile {output_dir}/Snakefile -j {cpu_per_job}'
        f.write(cmd + '\n')
    return


def check_chunk_dir_finish(output_dir):
    output_dir = pathlib.Path(output_dir)
    flag_path = output_dir / 'chunk_finished'
    if flag_path.exists():
        return

    path_not_finished = []
    for chunk_dir in output_dir.glob('*_chunk*'):
        if not (chunk_dir / 'finish').exists():
            path_not_finished.append(str(chunk_dir))
    if len(path_not_finished) > 0:
        path_str = '\n'.join(path_not_finished)
        raise ValueError('These chunk dirs have not finish successfully:\n' + path_str)
    else:
        with open(flag_path, 'w') as _:
            pass
    return



import time
import cooler
import numpy as np
from scipy.ndimage import convolve
from scipy.sparse import csr_matrix, save_npz, triu
from scipy.stats import zscore


def calc_diag_stats(E, n_dims):
    """Calculate cutoff, average, std, count of non-zero pixels of each diagonals of the E"""
    ave, std, top, count = np.zeros((4, n_dims), dtype=np.float32)
    for i in range(n_dims):
        tmp = E.diagonal(i)
        if tmp.size == 0:
            top[i] = 0
            ave[i] = 0
            std[i] = 0
            count[i] = 0
        else:
            cutoff = np.percentile(tmp, 99)
            tmp = np.where(tmp < cutoff, tmp, cutoff)
            top[i] = cutoff
            ave[i] = np.mean(tmp)
            std[i] = np.std(tmp)
            count[i] = np.sum(tmp > 0)
        # TODO smoothing
    return ave, std, top, count


def calculate_chrom_background_normalization(cell_url,
                                             chrom,
                                             resolution,
                                             output_prefix,
                                             dist=10050000,
                                             cap=5,
                                             pad=5,
                                             gap=2,
                                             min_cutoff=1e-6,
                                             log_e=False):
    """
    Compute the background for each chromosome in each cell

    Parameters
    ----------
    cell_url
    chrom
    resolution
    output_prefix
    dist
    cap
    pad
    gap
    min_cutoff
    log_e

    Returns
    -------
    E is the global diagonal normalized matrix
    T is the local background normalized version of E
    """
    cell_cool = cooler.Cooler(cell_url)
    # Load the cell imputed matrix as E
    E = triu(cell_cool.matrix(balance=False, sparse=True).fetch(chrom))
    E = E.astype(np.float32).toarray()

    # create an upper triangle mask
    mask = np.zeros(E.shape, dtype=bool)
    row, col = np.diag_indices(E.shape[0])
    mask[row, col] = True
    for i in range(1, dist // resolution + 1):
        mask[row[:-i], col[i:]] = True

    if log_e:
        # normalize E at log scale
        E[row, col] = 0
        for i in range(1, dist // resolution + 1):
            tmp = E.diagonal(i).copy()
            tmp_filter = (tmp > 0)
            tmp2 = tmp[tmp_filter]
            if len(tmp2) == 0:
                E[row[:-i], col[i:]] = 0
            else:
                tmp2 = zscore(np.log10(tmp2))
                tmp[~tmp_filter] = tmp2.min()
                tmp[tmp_filter] = tmp2.copy()
                E[row[:-i], col[i:]] = tmp.copy()
    else:
        # normalize E at linear scale
        # Calculate the diagonal stats of E
        n_dims = dist // resolution + pad + 1
        ave, std, top, count = calc_diag_stats(E, n_dims)
        # normalize E with the diagonal backgrounds
        idx = np.where(mask)
        tmp = E[idx]
        tmp = np.min([tmp, top[idx[1] - idx[0]]], axis=0)
        tmp = (tmp - ave[idx[1] - idx[0]]) / (std[idx[1] - idx[0]] + 1e-5)  # add a small value to prevent divide by 0
        tmp[count[idx[1] - idx[0]] < 100] = 0
        tmp[std[idx[1] - idx[0]] == 0] = 0
        tmp[tmp > cap] = cap
        tmp[tmp < -cap] = -cap
        E[idx] = tmp.copy()
    # print(f'Norm {time.time() - start_time:.3f}', E.dtype, tmp.dtype)

    # normalize E with the local backgrounds to generate T
    w = pad * 2 + 1
    kernel = np.ones((w, w), np.float32)
    kernel[(pad - gap):(pad + gap + 1), (pad - gap):(pad + gap + 1)] = 0
    kernel = kernel / np.sum(kernel)
    T = convolve(E, kernel, mode='mirror')
    E = csr_matrix(E)
    T = csr_matrix(T * mask)
    if min_cutoff > 0:
        # mask out small abs values
        E = E.multiply(np.abs(E) > min_cutoff)
        T = T.multiply(np.abs(T) > min_cutoff)
    T = E - T
    # print(f'Bkg {time.time() - start_time:.3f}', E.dtype, T.dtype)

    save_npz(f'{output_prefix}.E.npz', E)
    save_npz(f'{output_prefix}.T.npz', T)
    return



import sys
import pandas as pd

chrom_file = sys.argv[4]
chromsize = pd.read_csv(chrom_file, sep='\t', header=None, index_col=0).to_dict()[1]
res = 500000
size = [0]
for c in chromsize:
	size.append(int(chromsize[c] // res) + 1 + size[-1])

size = {xx:yy for xx,yy in zip(chromsize, size)}

infile = sys.argv[1]
outdir = sys.argv[2]
cell = sys.argv[3]
data = pd.read_csv(infile, sep='\t', header=None, index_col=None)
data = data[data[4]==data[5]]
data[4] = data[4].str.split('_', expand=True)[1]
data = data[data[4].isin(chromsize.keys())]
data[5] = data[4].map(size)
data[[0,1]] = data[[0,1]].values - data[5].values[:,None]
for c, tmp in data.groupby(by=4):
	tmp[[0,1,2]].astype(int).to_csv(f'{outdir}{c}/{cell}_{c}.txt', sep='\t', header=False, index=False)

tmp = data[4].unique()
for c in chromsize:
	if not (c in tmp):
		open(f'{outdir}{c}/{cell}_{c}.txt', 'w').close()



from setuptools import setup, find_packages

setup(
    name='schicluster',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Jingtian Zhou',
    author_email='jiz509@eng.ucsd.edu',
    packages=find_packages(),
    description='A package for single-cell Hi-C data clustering and downstream analysis.',
    long_description=open('README.md').read(),
    license='MIT',
    long_description_content_type='text/markdown',
    url='https://github.com/zhoujt1994/scHiCluster',
    include_package_data=True,
    install_requires=['numpy', 'scipy', 'scikit-learn', 'h5py', 'joblib', 'clodius',
                      'tables', 'cooler', 'pandas', 'statsmodels', 'rpy2', 'anndata', 'xarray'],
    package_data={
        '': ['*.txt', '*.tsv', '*.csv', '*.fa', '*Snakefile', '*ipynb', '*R']
    },
    entry_points={
        'console_scripts': ['hicluster=schicluster.__main__:main',
                            'hic-internal=schicluster._hicluster_internal:internal_main'],
    }
)



import time
import numpy as np
from scipy.sparse import csr_matrix, triu
import cooler
from ..cool.utilities import write_coo
import pandas as pd
import logging

"""
Matrix names
Q: The imputed, before normalization matrix.
"""


def read_chrom(cell_url, chrom):
    cool = cooler.Cooler(cell_url)
    matrix = triu(cool.matrix(balance=False, sparse=True).fetch(chrom))
    return matrix


def merge_cells_for_single_chromosome(cell_urls_path,
                                      chrom,
                                      output_prefix):
    cell_table = pd.read_csv(cell_urls_path, index_col=0)
    cell_urls = cell_table['cell_url'].tolist()
    n_cells = len(cell_urls)
    # get n_dims
    matrix = read_chrom(cell_urls[0], chrom)
    n_dims = matrix.shape[0]

    start_time = time.time()
    print('Merging Q (imputed, before normalization) matrix.')
    # initialize
    matrix_sum = csr_matrix((n_dims, n_dims), dtype=np.float32)
    for i, path in enumerate(cell_urls):
        matrix_sum += read_chrom(path, chrom)
    # normalize by total cell numbers
    matrix_sum.data /= n_cells
    write_coo(f'{output_prefix}.Q.hdf', matrix_sum, chunk_size=None)
    logging.debug(f'Merge {n_cells} cells took {time.time() - start_time:.0f} seconds')
    return



# python /gale/ddn/snm3C/humanPFC/code/generatematrix_cell.py --infile /gale/netapp/seq4/illumina_runs/ENTEx/Pool_BX_BY/mapping/output/TrCo_A9HOW_Plate1-1-H22/hic/TrCo_A9HOW_Plate1-1-H22-A13.3C.sorted_contacts_old.txt --outdir /gale/ddn/snm3C/humanPFC/cell_matrix/test/10kb_resolution/ --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes --split_file /gale/netapp/home/zhoujt/genome/hg19/hg19.chrsplit.bed --res 10000 --cell TrCo_A9HOW_Plate1-1-H22-A13
# python /gale/ddn/snm3C/humanPFC/code/generatematrix_cell.py --infile /gale/netapp/seq4/illumina_runs/ENTEx/Pool_BX_BY/mapping/output/TrCo_A9HOW_Plate1-1-H22/hic/TrCo_A9HOW_Plate1-1-H22-A13.3C.sorted_contacts_old.txt --outdir /gale/ddn/snm3C/humanPFC/cell_matrix/test/100kb_resolution/ --chrom_file /gale/netapp/home/zhoujt/genome/hg19/hg19.autosomal.chrom.sizes --res 100000 --cell TrCo_A9HOW_Plate1-1-H22-A13

import os
import argparse
import numpy as np
import pandas as pd
#from collections import Counter

def generatematrix_cell(infile, outdir, cell, res, chrom_file, 
					chr1=1, pos1=2, chr2=5, pos2=6, split_file=None, dist=2500):

	chrom = np.loadtxt(chrom_file, dtype=np.str)[:,0]

	# add p/q arm to split chromosomes
	if not split_file:
		chrom_split = chrom.copy() 
	else:
		splitbed = pd.read_csv(split_file, sep='\t', header=None, index_col=0)
		chrom_split = np.concatenate([[c+'p', c+'q'] if c in splitbed.index else [c] for c in chrom])

	# create a folder per chromosome, containing all cells
	for c in chrom_split:
		if c[:3]=='chr':
			os.makedirs(f'{outdir}{c}/', exist_ok=True)
		else:
			os.makedirs(f'{outdir}chr{c}/', exist_ok=True)

	data = pd.read_csv(infile, sep='\t', header=None)
	# make pos1 < pos2
	data.loc[data[pos1]>data[pos2], [pos1, pos2]] = data.loc[data[pos1]>data[pos2], [pos2, pos1]]
	# filter intra autosomal contacts with insertion size threshold
	data = data[(data[chr1]==data[chr2]) & (data[chr1].isin(chrom)) & (data[pos2] - data[pos1] > dist)]
	if split_file:
		splitfilter = data[chr1].isin(splitbed.index)
		datasplit = data[splitfilter]
		# end position of p and start position of q
		datasplit['p'] = splitbed.loc[datasplit[chr1], 1].values
		datasplit['q'] = splitbed.loc[datasplit[chr1], 2].values
		datasplit.loc[datasplit[pos1]<datasplit['p'], chr1] += 'p'
		datasplit.loc[datasplit[pos2]<datasplit['p'], chr2] += 'p'	
		datasplit.loc[datasplit[pos1]>datasplit['q'], chr1] += 'q'
		datasplit.loc[datasplit[pos2]>datasplit['q'], chr2] += 'q'
		# fully divide to ensure bins consistency between p and q
		datasplit.loc[datasplit[pos1]>datasplit['q'], pos1] -= datasplit['q'] // res * res
		datasplit.loc[datasplit[pos2]>datasplit['q'], pos2] -= datasplit['q'] // res * res
		data = pd.concat([datasplit[data.columns], data[~splitfilter]], axis=0)
	data[[pos1, pos2]] = data[[pos1,pos2]] // res
	data = data[(data[chr1]==data[chr2]) & (data[chr1].isin(chrom_split)) & (data[pos1]!=data[pos2])]
	# count contacts at each pixel
	data = data.groupby(by=[chr1, pos1, pos2])[chr2].count().reset_index()
	for c, tmp in data.groupby(by=chr1):
		if c[:3]=='chr':
			tmp[[pos1, pos2, chr2]].astype(int).to_csv(f'{outdir}{c}/{cell}_{c}.txt', sep='\t', header=False, index=False)
		else:
			tmp[[pos1, pos2, chr2]].astype(int).to_csv(f'{outdir}{c}/{cell}_chr{c}.txt', sep='\t', header=False, index=False)
	'''
	count = Counter()
	with open(infile, 'r') as fin:
		for line in fin:
			tmp = line.strip().split('\t')
			c1, c2, p1, p2 = tmp[chr1], tmp[chr2], int(tmp[pos1]), int(tmp[pos2])
			if p1>p2:
				p1, p2 = p2, p1
			if split_file:
				if c1 in splitbed.index:
					if p1 < splitbed.loc[c1,0]:
						c1 = c1 + 'p'
					elif p1 > splitbed.loc[c1,1]:
						c1 = c1 + 'q'
						p1 = p1 - splitbed.loc[c1,1]
				if c2 in splitbed.index:
					if p2 < splitbed.loc[c2,0]:
						c2 = c2 + 'p'
					elif p2 > splitbed.loc[c2,1]:
						c2 = c2 + 'q'
						p2 = p2 - splitbed.loc[c2,1]
			if (c1==c2) and (c1 in chrom_split) and (p2 - p1 >= dist):
				p1, p2 = p1 // res, p2 // res
				if p1!=p2:
					pos = '-'.join([c1, str(p1), str(p2)])
					count[pos] += 1
	fout = {c:open(outdir + 'chr' + c + '/' + cell + '_chr' + c + '.txt', 'w') for c in chrom_split}
	for key in count:
		tmp = key.split('-')
		fout[tmp[0]].write('{0}\t{1}\t{2}.0\n'.format(tmp[1], tmp[2], count[key]))
	for c in chrom:
		fout[c].close()
	'''
	return

'''
parser = argparse.ArgumentParser()
parser.add_argument('--infile', type=str, default=None, help='Path to the short format contact file')
parser.add_argument('--chr1', type=int, default=2, help='Column of fragment1 chromosome')
parser.add_argument('--pos1', type=int, default=3, help='Column of fragment1 position')
parser.add_argument('--chr2', type=int, default=6, help='Column of fragment2 chromosome')
parser.add_argument('--pos2', type=int, default=7, help='Column of fragment2 position')
parser.add_argument('--outdir', type=str, default=None, help='Output directory end with /')
parser.add_argument('--cell', type=str, default=None, help='Specific identifier of a cell')
parser.add_argument('--res', type=int, default=None, help='Bin size as integer to generate contact matrix')
parser.add_argument('--chrom_file', type=str, default=None, help='Path to the chromosome size files containing all chromosome to be analyzed as the first column') ## mm10 or hg38
parser.add_argument('--split_file', type=str, default=None, help='Path to the bed file containing all chromosomes need to split and one region per chromosome as splitting point')
parser.add_argument('--dist', type=int, default=2500, help='Minimum distance threshold of contacts to use')
opt = parser.parse_args()

generatematrix_cell(opt.infile, opt.outdir, opt.cell, opt.res, opt.chrom_file, 
				opt.chr1-1, opt.pos1-1, opt.chr2-1, opt.pos2-1, opt.split_file, opt.dist)
'''


import time
import numpy as np
from scipy.sparse import csr_matrix, diags, eye
from scipy.sparse.linalg import norm
from scipy.ndimage import gaussian_filter
import cooler
import logging
from ..cool import write_coo


def calc_sparsity(matrix):
    row, col = matrix.shape
    sparsity = matrix.nnz / row / col
    return sparsity


def random_walk_cpu(P, rp, tol):
    if rp == 1:
        return P

    _start_time = time.time()
    n_genes = P.shape[0]
    I = eye(n_genes, dtype=np.float32)
    Q = P.copy()
    for i in range(30):
        Q_new = P.dot(Q * (1 - rp) + rp * I)
        delta = norm(Q - Q_new)
        Q = Q_new.copy()
        sparsity = calc_sparsity(Q)
        _end_time = time.time()
        logging.debug(
            f'Iter {i + 1} takes {(_end_time - _start_time):.3f} seconds. '
            f'Loss: {delta:.3f}; Sparsity: {sparsity:.3f}', P.dtype, Q.dtype)
        if delta < tol:
            break
    return Q


def impute_chromosome(scool_url,
                      chrom,
                      resolution,
                      output_path,
                      logscale=False,
                      pad=1,
                      std=1,
                      rp=0.5,
                      tol=0.01,
                      window_size=500000000,
                      step_size=10000000,
                      output_dist=500000000,
                      min_cutoff=0):
    cell_cool = cooler.Cooler(scool_url)
    A = cell_cool.matrix(balance=False, sparse=True).fetch(chrom)
    n_bins = A.shape[0]
    ws = int(window_size // resolution)
    ss = int(step_size // resolution)

    # log transform
    if logscale:
        A.data = np.log2(A.data + 1)

    # Gaussian convolution and
    start_time = time.time()
    if pad > 0:
        # full matrix step
        A = gaussian_filter((A + A.T).astype(np.float32).toarray(), std, order=0, mode='mirror', truncate=pad)
        A = csr_matrix(A)
    else:
        A = A + A.T
    end_time = time.time()
    logging.debug(f'Convolution takes {end_time - start_time:.3f} seconds')

    # Remove diagonal before RWR
    A = A - diags(A.diagonal())

    # Random Walk with Restart
    start_time = time.time()
    if ws >= n_bins or rp == 1:
        B = A + diags((A.sum(axis=0).A.ravel() == 0).astype(int))
        d = diags(1 / B.sum(axis=0).A.ravel())
        P = d.dot(B).astype(np.float32)
        E = random_walk_cpu(P, rp, tol)
    else:
        # if the chromosome is too large, compute by chunks
        idx = (np.repeat(np.arange(ws), ws), np.tile(np.arange(ws), ws))
        idxfilter = (np.abs(idx[1] - idx[0]) < (output_dist // resolution + 1))
        idx = (idx[0][idxfilter], idx[1][idxfilter])
        # first filter
        idxfilter = ((idx[0] + idx[1]) < (ws + ss))
        idx1 = (idx[0][idxfilter], idx[1][idxfilter])
        mask1 = csr_matrix((np.ones(len(idx1[0])), (idx1[0], idx1[1])),
                           (ws, ws))
        # last filter
        idxfilter = ((idx[0] + idx[1]) >= (
                (n_bins - ws) // ss * 2 + 1) * ss + 3 * ws - 2 * n_bins)
        idx2 = (idx[0][idxfilter], idx[1][idxfilter])
        mask2 = csr_matrix((np.ones(len(idx2[0])), (idx2[0], idx2[1])),
                           (ws, ws))
        # center filter
        idxfilter = np.logical_and((idx[0] + idx[1]) < (ws + ss),
                                   (idx[0] + idx[1]) >= (ws - ss))
        idx0 = (idx[0][idxfilter], idx[1][idxfilter])
        mask0 = csr_matrix((np.ones(len(idx0[0])), (idx0[0], idx0[1])),
                           (ws, ws))

        start_time = time.time()
        E = csr_matrix(A.shape, dtype=np.float32)
        for ll in [x for x in range(0, n_bins - ws, ss)] + [n_bins - ws]:
            B = A[ll:(ll + ws), ll:(ll + ws)]
            B = B + diags((B.sum(axis=0).A.ravel() == 0).astype(int))
            d = diags(1 / B.sum(axis=0).A.ravel())
            P = d.dot(B).astype(np.float32)
            Etmp = random_walk_cpu(P, rp, tol)
            if ll == 0:
                E[ll:(ll + ws), ll:(ll + ws)] += Etmp.multiply(mask1)
            elif ll == (n_bins - ws):
                E[ll:(ll + ws), ll:(ll + ws)] += Etmp.multiply(mask2)
            else:
                E[ll:(ll + ws), ll:(ll + ws)] += Etmp.multiply(mask0)
    logging.debug(f'RWR takes {time.time() - start_time:.3f} seconds')

    # Normalize
    start_time = time.time()
    E += E.T
    d = E.sum(axis=0).A.ravel()
    d[d == 0] = 1
    b = diags(1 / np.sqrt(d))
    E = b.dot(E).dot(b)
    logging.debug(f'SQRTVC takes {time.time() - start_time:.3f} seconds')

    start_time = time.time()
    # mask the lower triangle of E
    # TODO This part is MEM intensive, the mask below can be combined with the chunk mask above
    idx = np.triu_indices(E.shape[0], 0)
    if (output_dist // resolution + 1) < n_bins:
        # longest distance filter mask
        idxfilter = ((idx[1] - idx[0]) < (output_dist // resolution + 1))
        idx = (idx[0][idxfilter], idx[1][idxfilter])
    mask = csr_matrix((np.ones(len(idx[0])), (idx[0], idx[1])),
                      E.shape,
                      dtype=np.float32)
    E = E.tocsr().multiply(mask)
    logging.debug(f'Filter takes {time.time() - start_time:.3f} seconds')

    # TODO put this part inside RWR, before normalize
    # min_cutoff = tol/
    # Make values < min_cutoff to 0
    if min_cutoff > 0:
        s_before = calc_sparsity(E)
        E = E.multiply(E > min_cutoff)
        s_after = calc_sparsity(E)
        logging.debug(f'Mask values smaller than {min_cutoff}. Sparsity before {s_before:.3f}, after {s_after:.3f}')

    # save to file
    write_coo(output_path, E)
    return
